(******************************************************************************)
(*                                                                            *)
(*                                    Menhir                                  *)
(*                                                                            *)
(*   Copyright Inria. All rights reserved. This file is distributed under     *)
(*   the terms of the GNU Library General Public License version 2, with a    *)
(*   special exception on linking, as described in the file LICENSE.          *)
(*                                                                            *)
(******************************************************************************)

open TableFormat
open GLRAPI

(**This module decodes the tables generated by the GLR back-end
   and provides an API that is suitable for use by the GLR engine. *)

module Make (T : TABLES) : DATA
  with type state = int
   and type token = T.token
   and type nonterminal = int
   and type semv = Obj.t
   and type production = int
